use std::cell::RefCell;
use std::rc::Weak;

mod cpu;
mod mappers;
mod ppu;
mod rom;
pub mod system;

pub trait Clock {
    /// Run a clock cycle
    fn cycle(&mut self);
}

pub trait MemoryBus {
    /// Gets a byte from the memory bus.
    ///
    /// # Arguments
    ///  * `addr` - The address of the byte
    fn read_byte(&self, addr: u16) -> u8;

    /// Sets a byte in the memory bus
    ///
    /// # Arguments
    ///  * `addr` - The address of the byte
    ///  * `value` - The value to set
    fn write_byte(&mut self, addr: u16, value: u8);
}

pub trait Observable<Key, Value> {
    /// Registers an observer to be notified on changes.
    ///
    /// # Arguments
    ///  * observer - An observer implementation to register
    fn register_observer(&mut self, observer: Weak<RefCell<dyn Observer<Key, Value>>>);
}

pub trait Observer<Key, Value> {
    /// Called when a value was updated in an observable.
    ///
    /// # Arguments
    ///  * key - The key of the value
    ///  * value - The updated value
    fn notify_updated(&mut self, key: Key, value: Value);
}
