use crate::mappers::Mapper;
use crate::rom::NametableMirroring;
use crate::MemoryBus;
use simplelog::warn;

const NAMETABLE_SIZE: usize = 0x400;
const PALETTE_TABLE_SIZE: usize = 0x20;
const VRAM_SIZE: usize = 0x4000;

pub struct PpuMemoryBus {
    /// The ROM mapper. Only present after a ROM has been "inserted" in the system.
    mapper: Option<Box<dyn Mapper>>,

    /// The first nametable.
    nametable_0: [u8; NAMETABLE_SIZE],

    /// The second nametable.
    nametable_1: [u8; NAMETABLE_SIZE],

    /// The palette table.
    palette: [u8; PALETTE_TABLE_SIZE],
}

impl PpuMemoryBus {
    pub fn new() -> Self {
        Self {
            mapper: None,
            nametable_0: [0u8; NAMETABLE_SIZE],
            nametable_1: [0u8; NAMETABLE_SIZE],
            palette: [0u8; PALETTE_TABLE_SIZE],
        }
    }

    fn read_nametable(&self, addr: u16) -> u8 {
        assert!(addr >= 0x2000);
        assert!(addr < 0x3000);

        let nametable_addr = (addr - 0x2000) as usize;
        let mirroring = match self.mapper.as_ref() {
            None => NametableMirroring::Horizontal,
            Some(mapper) => mapper.are_nametables_mirrored(),
        };

        if nametable_addr < NAMETABLE_SIZE {
            self.nametable_0[nametable_addr]
        } else if nametable_addr < NAMETABLE_SIZE * 2 {
            if mirroring == NametableMirroring::Horizontal {
                self.nametable_0[nametable_addr - NAMETABLE_SIZE]
            } else {
                self.nametable_1[nametable_addr - NAMETABLE_SIZE]
            }
        } else if nametable_addr < NAMETABLE_SIZE * 3 {
            if mirroring == NametableMirroring::Horizontal {
                self.nametable_1[nametable_addr - NAMETABLE_SIZE * 2]
            } else {
                self.nametable_0[nametable_addr - NAMETABLE_SIZE * 2]
            }
        } else {
            self.nametable_1[nametable_addr - NAMETABLE_SIZE * 3]
        }
    }

    fn write_nametable(&mut self, addr: u16, data: u8) {
        assert!(addr >= 0x2000);
        assert!(addr < 0x3000);

        let nametable_addr = (addr - 0x2000) as usize;
        let mirroring = match self.mapper.as_ref() {
            None => NametableMirroring::Horizontal,
            Some(mapper) => mapper.are_nametables_mirrored(),
        };

        let nametable = if nametable_addr < NAMETABLE_SIZE {
            self.nametable_0.as_mut_slice()
        } else if nametable_addr < NAMETABLE_SIZE * 2 {
            if mirroring == NametableMirroring::Horizontal {
                self.nametable_0.as_mut_slice()
            } else {
                self.nametable_1.as_mut_slice()
            }
        } else if nametable_addr < NAMETABLE_SIZE * 3 {
            if mirroring == NametableMirroring::Horizontal {
                self.nametable_1.as_mut_slice()
            } else {
                self.nametable_0.as_mut_slice()
            }
        } else {
            self.nametable_1.as_mut_slice()
        };

        let nametable_addr = addr as usize % NAMETABLE_SIZE;
        nametable[nametable_addr] = data;
    }
}

impl MemoryBus for PpuMemoryBus {
    fn read_byte(&self, addr: u16) -> u8 {
        assert!(addr < VRAM_SIZE as u16);

        if addr < 0x2000 {
            if self.mapper.is_none() {
                warn!(
                    "Tried to read PPU memory address {} before setting a mapper",
                    addr
                );
                return 0;
            }

            let mapper = self.mapper.as_ref().unwrap();
            mapper.get_chr_byte(addr)
        } else if addr < 0x3000 {
            self.read_nametable(addr)
        } else if addr < 0x3F00 {
            warn!("Read from unused PPU address space: {:#04x}", addr);
            0
        } else {
            let palette_addr = (addr as usize - 0x3F00) % PALETTE_TABLE_SIZE;
            self.palette[palette_addr]
        }
    }

    fn write_byte(&mut self, addr: u16, value: u8) {
        assert!(addr < VRAM_SIZE as u16);

        if addr < 0x2000 {
            warn!("Unsupported write to CHR ROM at address {:#04x}", addr);
        } else if addr < 0x3000 {
            self.write_nametable(addr, value);
        } else if addr < 0x3F00 {
            warn!("Write to unused PPU address space: {:#04x}", addr);
        } else {
            let palette_addr = (addr as usize - 0x3F00) % PALETTE_TABLE_SIZE;
            self.palette[palette_addr] = value;
        }
    }
}
