use crate::rom::ines::INesHeader;
use simplelog::{debug, info};
use std::fmt::{Display, Formatter};
use std::fs;

mod ines;

pub struct Rom {
    /// The size of the program RAM in bytes
    pub prg_ram_size: usize,
    
    /// The size of the non-volatile program RAM in bytes
    pub prg_nvram_size: usize,
    
    /// The size of the program ROM in bytes
    pub prg_rom_size: usize,

    /// The program ROM
    pub prg_rom: Box<[u8]>,
    
    /// The size of the character RAM
    pub chr_ram_size: usize,
    
    /// The size of the character RAM (non-volatile)
    pub chr_nvram_size: usize,

    /// The size of the character ROM in bytes
    pub chr_rom_size: usize,

    /// The character ROM
    pub chr_rom: Box<[u8]>,
    
    /// The mirroring of the nametable
    pub nametable_mirroring: NametableMirroring,
    
    /// The timing of the CPU, also the region of the ROM
    pub cpu_timing: CpuTiming,

    /// The number of the mapper
    pub mapper: u16,

    /// The number of the sub-mapper
    /// Not supported for now
    pub sub_mapper: u8
}

#[derive(Debug, Copy, Clone, Eq, PartialEq)]
pub enum NametableMirroring {
    Horizontal,
    Vertical,
}

#[derive(Debug, Copy, Clone)]
pub enum CpuTiming {
    Ntsc,
    Multiple,
    Pal,
    Dendy
}

pub trait RomLoader {
    /// Checks if a header is for this file format.
    ///
    /// # Arguments
    /// * `header` - The content of the header
    ///
    /// # Returns
    /// A boolean indicating if the header matches the ROM format.
    fn format_matches(header: &[u8]) -> bool;

    /// Load a ROM file.
    ///
    /// # Arguments
    ///  * `header` - The header of the ROM
    ///  * `file_data` - The data from the file
    ///
    /// # Returns
    /// The ROM that was loaded
    fn load(header: &[u8], file_data: &[u8]) -> Result<Rom, RomReadError>;
}

impl Rom {
    pub fn read(path: &str) -> Result<Rom, RomReadError> {
        info!("ROM - Reading file from {}", path);

        let content = match fs::read(path) {
            Ok(content) => content,
            Err(error) => return Err(RomReadError::Io(error)),
        };

        if content.len() < 16 {
            return Err(RomReadError::MissingHeader)
        }

        let header = &content[0..16];
        if !INesHeader::format_matches(header) {
            return Err(RomReadError::FormatNotSupported)
        }

        info!("ROM - Matched format iNes 2.0");
        
        let rom = INesHeader::load(header, &content[16..])?;
        
        debug!("ROM - PRG ROM: {} bytes, CHR ROM: {} bytes", rom.prg_rom_size, rom.chr_rom_size);
        info!("ROM - Loading successful");

        Ok(rom)
    }
}

#[derive(Debug)]
pub enum RomReadError {
    FormatNotSupported,
    Io(std::io::Error),
    InvalidHeader(String),
    MissingHeader,
    RomLengthMismatch(usize, usize),
    UnsupportedMapper(u8)
}

impl Display for RomReadError {
    fn fmt(&self, f: &mut Formatter<'_>) -> std::fmt::Result {
        match *self {
            RomReadError::FormatNotSupported => write!(f, "ROM didn't match any supported format"),
            RomReadError::Io(ref err) => write!(f, "IO error: {}", err),
            RomReadError::InvalidHeader(ref message) => write!(f, "Invalid ROM header: {}", message),
            RomReadError::MissingHeader => write!(f, "File was too short to contain a header"),
            RomReadError::RomLengthMismatch(required_length, actual_length) =>
                write!(f, "ROM didn't contain enough data according to the header; required {} bytes but contained {} bytes", required_length, actual_length),
            RomReadError::UnsupportedMapper(mapper) => write!(f, "Unsupported mapper: {}", mapper),
        }
    }
}
